/*
 * Domain Bridge Control
 *
 * Copyright 2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * --------------------------------------------------------------------------
 * $Id: dbctrl.c,v 1.2 2003/03/29 12:21:10 csmith Exp $
 * --------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_memlst.h>
#include <phlib/xl_malloc.h>

#define GW_INTERNAL_FLAGS 1

#include <gwneng.h>
#include <gwnevent.h>
#include <gwxlapi.h>
#include <goldreq.h>
#include <bbquery.h>
#include <gwgbprm.h>
#include <gwlog.h>
#include <gwapi.h>
#include <gwmsg.h>

#include "dbctrl.h"
#include "dbcom.h"
#include "pdbcom.h"
#include "dbstate.h"
#include "dbridge.h"
#include "version.h"

extern DomainNode *i_dom_array; /* Created in dbridge.c */

/* 
 * Const strings.... 
 */
const char SRV_DESC[] = "Goldwater Domain Bridge - (c)2002 netFluid Technology";
const char REVISION[] = DOM_VERSION;

const char READ_ACK[] = "PASS";

const char CRLF    [] = "\r\n";
const char CRLFCRLF[] = "\r\n\r\n";

long   msg_id_counter = 0;
 
#define MAXSERVERS    2
#define WR_BLOCK_SIZE 4096

NetService NServers[ MAXSERVERS ];

/* Prototype functions 
 */
Action read_handler ( int, void ** );
Action event_handler( EventMsg *, void ** );

/* Private Prototypes 
 */
static RetValue dispatch_message     ( DomSession *sc );
static RetValue make_dead_message    ( DomSession *sc );
static void     cleanup_message_chain( DomSession *sc );
static void     clean_dead_message   ( DomSession *sc );

/* ----------------------------------------------------------------------- */

RetValue
init_db_network()
{
  NetService ns;

  ns.port          = 0;
  ns.ipaddr        = "";
  ns.desc          = SRV_DESC;
  ns.svcidstr      = SERVICE;
  ns.read_handler  = read_handler;
  ns.event_handler = event_handler;

  if( gwne_netcom_init( NServers, 0, MAXSERVERS, WR_BLOCK_SIZE ) == Failure )
    return Failure;
  if( gwne_netcom_add_handler( &ns ) == Failure ) return Failure;

  return Success;
}

/* ----------------------------------------------------------------------- */
/*
 * Read handler is controlled by several states.
 */
Action
read_handler( int fd, void ** private )
{
  Action  ret = ENABLE_READ;
  char    buf[ DBR_PAYLOAD_IX ];
  int     n = 0;
  DomSession *sc = (DomSession*)(*private);
  
  if( *private == 0 ) return MDL_ERROR;

  switch( sc->phase ) {

  case Idle:
    /* This is the catch all 'black hole' read.
     * It is just used to detect when a connection has gone down - in which
     * case we'll get a read error.
     */
    if( (n = read( fd, buf, sizeof(buf)-1 )) <= 0 ) 
      return FREE_SESSION; /* Connection Probably Died */
    break;
    
  case ReadACK: /* ACK from request/result send */
    if( (n = read( fd, buf, sizeof(buf) )) <= 0 ) 
      return FREE_SESSION; /* Connection Probably Died */

    if( decode_ack( sc, buf ) != GWSuccess ) {
      /* Error was returned */
      gw_logf( LOG_ERROR, "Received error %d:%d", sc->status.result, sc->status.error );
   
      /* Report to client that a timeout has occured when we clean up the
       * dead messages
       */
      sc->current_dm->ret_error_on_cleanup = sc->status.error;
    }
    else gw_logf( LOG_INTDBG, "netserver recieved ack %d:%d", sc->status.result, sc->status.error );

    /* Move message into dead_message_chain for safe cleanup
     */
    make_dead_message( sc );
    /* sc->current_dm is now NULL */

    sc->phase = Idle;
    xlapi_remove_event( EV_TIME, EUT_ACK_TIMEOUT, 0 ); 

    /* Release the current message as it's been sent.
     */
    break;
  }

  return ret;
}

/* ----------------------------------------------------------------------- */

Action
event_handler( EventMsg *msg,  void ** private )
{
  Action  ret  = FINISHED;
  int     u_ev;
  DomSession    *sc = (DomSession*)(*private);
  DomainNode    *dn     = NULL;
  
  switch( msg->event ) {
    
  case EV_NEW_SESSION: /* Called when a new session is started */
    {
      DomSession  *ns;

      /* *private contains DOMAINID of the domain that we're connected to
       * and not a ptr to private session WS memory.
       * But that won't last long!
       * We're going to overwrite this *private data with a memory ptr
       * in just a few lines...... :o)
       */
      
      /* Create session memory */ 
      if( (ns=(DomSession*)xl_malloc(sizeof(DomSession)) ) != NULL ) {
 	
	ns->phase       = Idle;
	ns->domlist_idx = (int)*private;  /* private starts off as DOMAINID!! */
	ns->pending_count = 0;

	WrapCStringLen( ns->headCS, "EMPTY STRING", 0 ); /* Initialise CString */

	*private  = (void*)ns;  /* And now *private is the ptr to our Session*/

        gw_logf( LOG_INTDBG, "START NEW SESSION" );
	ret = ENABLE_READ;
      }
    }
    break;
  
  case EV_END_SESSION:
    {
      /* Invoked by the engine to clean up and exit a session.  
       * NO WRITES or events must be made we effectively no longer have a 
       * connection to the remote client/server.
       * 
       * Private space must be cleared here - Hence no further processing for
       * this session is possible once this function is exited.
       */
      BBDom *domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);
  
      /* sc->domlist_idx IS the domain id....
	   */
      gw_logf( LOG_AUDIT, "Connection lost with domain %s",
	       domarray[ sc->domlist_idx - 1 ].domain );

      domarray[ sc->domlist_idx - 1 ].available = 0; /* Mark as unavaliable */
  
      free_cstring( &sc->headCS );
  
      cleanup_message_chain( sc );
      clean_dead_message( sc );
  
      xl_free( sc ); /* Drop session memory */
      ret = TERMINATE_SESSION;    /* WE MUST RESPOND WITH THIS!  MANDATORY! */
    }
    break;
    
  case EV_WRDONE:
    /* There are no writes pending, and so it is safe to cleanup the dead
     * messages
     */
    clean_dead_message( sc );
    break;

  case EV_TIME:
    u_ev = msg->user_data; /*Get user assigned event code (from EV_USER_BASE)*/
    
    switch( u_ev ) {
    case EUT_ACK_TIMEOUT:
      
      /* Report to client that a timeout has occured when we clean up the
       * dead messages
       */
      if( sc->current_dm != NULL ) {
	sc->current_dm->ret_error_on_cleanup = GWE_TIMEOUT;

	/* Move message into dead_message_chain for safe cleanup */
	make_dead_message( sc );
	/* sc->current_dm is now NULL */
      }

      /* Remove some events, just in case */
      xlapi_remove_event( EV_USER_EVENT, EU_MSG_PENDING, 0 );
  
      sc->phase = Idle;
      break;
    }
    break;

    
  case EV_USER_EVENT:
    u_ev = msg->user_data;/*Get user assigned event code (from EV_USER_BASE)*/
    
    switch( u_ev ) {
      /* When we want to end our session NICELY we register EU_MY_SESSION_QUIT
       * This will gracefully wait for all writes to complete and perform
       * the tidy-ups that are required.
       */
    case EU_MY_SESSION_QUIT: /* Because of this code - DO NOT do a write API! */
      if( xlapi_write_done() == 0 ) {
	/* Here wait for all critical tidyups to be done.... */
        ret = FREE_SESSION; /* Indicate we want to get out */
      } else {
	xlapi_register_event( EV_USER_EVENT, EU_MY_SESSION_QUIT );
      }
      break;

    case EU_DEAD_MESSAGES:
      /* Keep trying to remove dead messages - this can only happen when we
       * have no writes pending.
       */
      dn = &i_dom_array[ sc->domlist_idx ];

      if( xl_List_get_size( dn->dead_message_chain ) ) {
        if( xlapi_write_done() == 0 ) {
	  clean_dead_message( sc );
        } else {
	  xlapi_register_event( EV_USER_EVENT, EU_DEAD_MESSAGES );
	}
      }
      break;

      /* An event from 'control' saying that a message has been added to the
       * message_chain, and should be dispatched.
       */
    case EU_NEW_MESSAGE:
      /* Pretend we've registered a MSG_PENDING event because we're going to
       * drop straight into it's handler.
       */
      sc->pending_count++; 
      
    case EU_MSG_PENDING:
      /* This event keeps repeating whilst we have messages in the msg_chain
       * and our state is NOT 'idle'.
       */

      /* Check for timed-out messages HERE *before* trying to send them! 
       *
       * dispatch_message should do the timeout checking when it takes the
       * msg off the head of the queue.  If it has timed out, it will need to
       * send a timeout response to the originating client (gw_send_reply).
       * This will also clean up the GWMsgCtrl struct and associated memory and
       * spool files.
       */

      /* If we're in the Idle state then we're "Okay To Go" with a new message.
       * Else, restore event so we get to check again!
       */
      if( sc->phase == Idle ) {
		
	if( dispatch_message( sc ) == Failure ) {
	  gw_logf( LOG_SYSWARN, "Failed to dispatch message" );
	  
	  if( make_dead_message( sc ) == Failure ) {
	    /* Ooooh Nooooooo! */
	    gw_logf( LOG_SERIOUS, "Could not mark message as dead" );

	    ret = FREE_SESSION; /* Indicate we want to get out RIGHT NOW! */
	  }
	  break;
	} else {

	  sc->pending_count--;  /* Dec count */

	  xlapi_register_time_event( EUT_ACK_TIMEOUT, 5 );
	}
      }

      /* If we've only got one message pending, then generate a pending event
       * (this is either the first message we've received and thus come in
       *  through EU_NEW_MESSAGE, or we've received a EU_MSG_PENDING event
       *  which needs to be refreshed).
       * If there are more than one msgs pending, then we must only generate
       * one msg_pending event per session, so DO NOT allow EU_NEW_MESSAGE
       * events to fall through this code and register an event (unless we've
       * only got one in the queue, as above).
       */
      if(  sc->pending_count == 1 || 
	   (sc->pending_count >  1 && u_ev == EU_MSG_PENDING ) ) {

	xlapi_register_event( EV_USER_EVENT, EU_MSG_PENDING );
      }

      break;
    }
  }
  
  return ret;
}

/* ----------------------------------------------------------------------- */
/*
 * This cleans up any messages that are in the ACTIVE message chain
 */
static void
cleanup_message_chain( DomSession *sc )
{
  DomainMessage *dm = NULL;
  DomainNode    *dn = NULL;

  /* Remove messages from message_chain and tidy up GWMsgCtrl data
   */
  dn = &i_dom_array[ sc->domlist_idx ];
  while( xl_List_get_first( dn->message_chain, 
                            (unsigned char**)&dm ) == Success ){
    gwp_cleanup_ctrl( &dm->rctrl );
    xl_List_delete_node( dn->message_chain );
  }
} 

/* ----------------------------------------------------------------------- */
/*
 * Moves the sc->current_dm DomainMessage into the dead_message_chain and
 * deletes it from the current domains active message chain.
 * It does this by *searching* the active list for the message that has the
 * same msgid as the current_dm messsage.  If it finds it, it does the
 * transfer.
 * MOST of the time this will be the first node in the list, but the search
 * is attempted just in case it is not.
 *
 * sc->current_dm is set to NULL if the transfer is made.
 */
static RetValue
make_dead_message( DomSession *sc )
{
  DomainNode    *dn    = NULL;
  DomainMessage *dm    = NULL;
  int            found = 0;
 
  if( sc->current_dm == NULL ) {
    gw_logf( LOG_ERROR, "make_dead_message called to transfer a NULL DomainMessage" );
    return Failure;
  }

  /* Track our domain control structure 
   */
  dn = &i_dom_array[ sc->domlist_idx ];

  /* Position at start of list
   */
  if( xl_List_get_first( dn->message_chain, (unsigned char**)&dm ) == Success ){
    /* Search for current message MSGID in the active message chain 
     */
    do {
      if( gwuuid_compare( dm->rctrl.msgid, sc->current_dm->rctrl.msgid ) == 0 ) {
	found = 1;
	break;
      }
    } while( xl_List_get_first( dn->message_chain, (unsigned char**)&dm ) == Success );
  } 

  if( found ) {
    /* Transfer DomainMessage into dead list for cleaning later
     */
    if( xl_List_transfer_node( dn->dead_message_chain, dn->message_chain, ListEnd ) == Failure ) return Failure;

    if( xl_List_get_size( dn->dead_message_chain ) == 1 ) {
      /* Only register one DEAD_MESSAGES event
       */
      xlapi_register_event( EV_USER_EVENT, EU_DEAD_MESSAGES );
    }

    if( gw_log_at_level(LOG_INTDBG) ) {
      char uuidbuf[64];
      uuid_unparse( dm->rctrl.msgid, uuidbuf );
      gw_logf( LOG_INTDBG, "Message %s now dead (dead list %d contains %d msgs)", uuidbuf, sc->domlist_idx, xl_List_get_size( dn->dead_message_chain ) );
    }
    sc->current_dm = NULL;
  }

  return Success;
}

/* ----------------------------------------------------------------------- */
/*
 * This cleans up any messages that are in the DEAD message chain
 */
static void
clean_dead_message( DomSession *sc )
{
  DomainNode    *dn     = NULL;
  DomainMessage *dm     = NULL;
  GWMsgCtrl      gwmold = {0};
  int            dom_request = 0;

  /* Track our domain control structure 
   */
  dn = &i_dom_array[ sc->domlist_idx ];

  gw_logf( LOG_INTDBG, "Cleaning %d msgs from dead list [%d]", xl_List_get_size( dn->dead_message_chain ), sc->domlist_idx );

  while( xl_List_get_first( dn->dead_message_chain, 
                            (unsigned char**)&dm ) == Success ) {

    dom_request = (dm->rctrl.msgtype == GWM_DOM_REQUEST) ? 1 : 0;

    gw_logf( LOG_INTDBG, "Update SVCCount in domain %d", dm->rctrl.target_dom );

    /* If the dead message requires an 'error' to be returned to the
     * message originator, do so before attempting a cleanup.
     *
     * WE CAN ONLY DO THIS IF WE ARE PROCESSING A DOMAIN REQUEST!
     * (because only then are we talking to a client.  If we're doing a
     *  Domain reply, we're getting async data from a server!)
     */
    if( dom_request && dm->ret_error_on_cleanup > 0 ) {

      /* gwmold <- req_ctrl <- message.rctrl */
      __gw_set_requestCtrl( &dm->rctrl, &gwmold );

      gwerror = dm->ret_error_on_cleanup;

      gw_send_reply( GWDomFailure, "", 0 );

      /* message.rctrl <- req_ctrl <- gwmold */
      __gw_set_requestCtrl( &gwmold, &dm->rctrl );
    }

    /* Finish the request.
     * This will update the server and service counters of the originating
     * domain if that's who we are (as we'd expect with a normal request).
     * 
     * DO NOT update any counters if were the remote domain!  The called
     * server does that itself!
     *
     * ie update counters iff we're a DOM_REQUEST and NOT an error
     */
    __gw_complete_request_explicit( &dm->rctrl, 
				    (dom_request && 
				     dm->ret_error_on_cleanup == 0 ) ? 1 : 0, 
				    dm->rctrl.service, dm->rctrl.target_dom );

    gw_logf( LOG_INTDBG, "Message Cleaned" );

    xl_List_delete_node( dn->dead_message_chain );
  }
      
  gw_logf( LOG_INTDBG, "Cleaned dead messages" );
}

/* ----------------------------------------------------------------------- */
/* Takes the next message queued for the sessions connected domain and
 * dispatches it via the xlapi_write API.
 *
 * The type of message header generated depends on the message type (be it a
 * request or a response message).
 */
RetValue
dispatch_message( DomSession *sc )
{
  DomainNode    *dn    = NULL;
  DomainMessage *dm    = NULL;

  /* Track our domain control structure 
   */
  dn = &i_dom_array[ sc->domlist_idx ];

  gw_logf( LOG_INTDBG, "Dispatch message from domain %d", sc->domlist_idx );

  if( xl_List_get_first( dn->message_chain, (unsigned char**)&dm ) == Failure ){
    /* Huh? How can this happen? */
    return Failure;
  }

  /* Keep a pointer to the current DM just in case ACK needs it */
  sc->current_dm = dm;
  sc->current_dm->ret_error_on_cleanup = 0;

  /* Decide what type of message we have so we can send the correct header 
   */
  switch( dm->rctrl.msgtype ) {
  case GWM_DOM_REQUEST:
    break;

  case GWM_DOM_REPLY:
    break;

  default:
    return Failure;
    break;
  }

  /* Create DomainMessage header of version 001 - placed in sc->headCS */
  if( gdbm_create_header( DBM_VERSION_001, dm, dn->domainid, sc ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Cannot create domain message - %s", gw_errorstr(gwerror) );    
    return Failure;
  }

  gw_logf( LOG_INTDBG, "Header generated of length %d", CSLen(sc->headCS) );

  gw_logf( LOG_INTDBG, "Send Header..." );

  /* Write Header */
  xlapi_write( &sc->headCS );
   
  /* Write Authoristion Data */
  xlapi_write_data( "01234567890123456", 16 ); /* MD5 BINARY (128bit/16 byte) */

  /* Write Message */
  gw_logf( LOG_INTDBG, "Writing %d of data to netServer", CSLen(dm->rctrl.cache) );
  xlapi_write( &dm->rctrl.cache ); /* data is in a CString */

  sc->phase = ReadACK;

  return Success;
}

/* ----------------------------------------------------------------------- */
/* end */
