/*
 * ---------------------------------------------------------------------------
 * Goldwater Distributed Web Architecture.
 *
 * Tool for compiling Goldwater Configuration into a binary tree.
 *
 * Copyright 1996-2002 netFluid Technology Limited
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF NETFLUID TECHNOLOGY LIMITED
 *
 * The copyright notice above does not evidence any actual or intended 
 * publication of such source code.    
 *
 * $Revision: 1.3 $  $Date: 2003/03/29 12:21:11 $
 * 
 * --------------------------------------------------------------------------
 * $Log: conftree.c,v $
 * Revision 1.3  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.5  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.4  2002/03/17 16:15:26  chris
 * Do not overwrite GWCONFIG if it already exists.
 *
 * ---------------------------------------------------------------------------
 */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_linklst.h>
#include <phlib/xl_memlst.h>
#include <phlib/xl_memory.h>
#include <phlib/xl_mmap.h>

#include "conftree.h"

#define SEP '/'

#ifdef DEBUG
# define IF_DEBUG if(1)
#else
# define IF_DEBUG if(0)
#endif

#define BOUNDARY 4

static LIST * ConfigList   = 0;
static int    flat_r_count = 0;

/* Prototypes
 */
static LIST *create_tree  ( LIST *stree, CString *branchCS  );
static void *find_node    ( LIST *, CString *               );
static int   create_leaf  ( LIST *, CString *, char *, int32, int);
static void  traverse_tree( LIST *tree, int tab );
static int   flatten_tree_r( LIST *tree, CString *memoryCS );

#ifdef WILD_MATCH
static int   wild_compare ( const CString* , const CString* );
#endif

RetValue RegDataChar( CString *treeCS, const char *path, char **data, long *len );

/* ----------------------------------------------------------------------- */

static RetValue
destroy_tree( LIST **tree )
{
  TreeNode *data;

  if( xl_List_get_first( *tree, (uchar**)&data ) == Success ) {
    do { 
      if( data->ttype == T_BRANCH ) destroy_tree( &(data->stree) );
    } while ( xl_List_get_next( *tree, (uchar**)&data ) == Success );
  }
  xl_List_end( tree );

  return Success;
}


/* ----------------------------------------------------------------------- */

RetValue
end_tree()
{
  return destroy_tree( &ConfigList );
}

/* ----------------------------------------------------------------------- */

RetValue
start_tree()
{
  RetValue ret = Success;

  if( ConfigList != 0 ) /* Destroy an existing tree */
    destroy_tree( &ConfigList );

  xl_List_init( &ConfigList, 1 ); /* Size is updated later */

  return ret;
}

/* ------------------------------------------------------------------------ */
/*
 * Adds a branch to the registry tree
 */
RetValue
reg_add_branch( char *branch )
{
  CString   tmpCS = CSInit;
  RetValue  ret   = Success;
  LIST     *stree = 0;
  TreeNode *tnode = 0;

  if( ConfigList == 0 ) return Failure;

  WrapCString( tmpCS, branch );
  stree = traverse_branch( ConfigList, &tmpCS, &tnode );

  /* TODO: Error properly */  
  if( stree == 0 ) { printf( "Cross leaf\n" ); return Failure; } /* branch crosses a leaf node! */

  if( tnode == 0 ) { /* Right.  Create subtree from returned tree stem */
    create_tree( stree, &tmpCS );
  }
    
  return ret;
}

/* ------------------------------------------------------------------------ */
/*
 * This call will fail if the registry leaf already exists
 */
RetValue
reg_add_leaf( char *branch, char *data, int len, int pos )
{
  CString  tmpCS = CSInit;
  RetValue ret   = Success;

  if( ConfigList == 0 ) return Failure;

  WrapCString( tmpCS, branch );
  if( create_leaf( ConfigList, &tmpCS, data, len, pos ) != 0 ) ret = Failure;
  
  return ret;
}

/* ------------------------------------------------------------------------ */
/*
 * Create a brand new tree (or subtree off an existing branch) 
 * Any part of the subtree being created is assumed NOT to exist.
 *
 * Returns last branch chain created.
 */
static LIST *
create_tree( LIST *stree, CString *branchCS )
{
  CString   tmpCS = CSInit;
  TreeNode *node;
  LIST     *ctree = stree;
  int       nsize;

  /* Create each node */
  while( chop_line( &tmpCS, branchCS, SEP ) == Success ) {
    
    if( CSLen( tmpCS ) == 0 ) continue; /* Skip empty segments */
    
    nsize = sizeof( TreeNode ) + CSLen( tmpCS ) + 1; /* NO DATA HERE */
    
    xl_List_set_datasize( ctree, nsize );         /* Specify space required  */
    xl_List_insert_space( ctree, ListEnd );       /* Create node with space  */
    node = (TreeNode*)xl_List_get_current_data( ctree ); /* get ptr to space */
    
    /* Start filling in details */
    node->ttype    = T_BRANCH;
    node->data     = 0;        /* No data */
    node->name     = (char*)node + sizeof( TreeNode );

    strncpy( node->name, CSStart(tmpCS), CSLen(tmpCS) );
    node->name[ CSLen(tmpCS) ] = '\0';
    node->namelen = CSLen(tmpCS);
    
    /* Now we need to create a new branch for further things */
    xl_List_init( &(node->stree), 1 ); /* Size updated later */
    ctree = node->stree;
  }
  
  return ctree;
}

/* ------------------------------------------------------------------------ */
/*
 * Create a leaf on a branch
 * 
 * Takes  Data Type, Pointer to data, length of data
 * Returns leaf_error
 *  0 = okay
 * -1 = branch crosses a leaf node
 * -2 = Leaf already exists
 */
static int
create_leaf( LIST *stree, CString *pathCS, char *data, int32 len, int pos )
{
  CString   branchCS = CSInit;
  CString   leafCS   = CSInit;
  TreeNode *node     = 0;
  ListPos   np       = pos;
  LIST     *ctree    = stree;
  int32     basesize, nsize;

  /* Keep a copy of the branch */
  MirrorCString( branchCS, *pathCS );
  chop_line_last( &leafCS, &branchCS, '/' ); /* Chop off leaf */

  /* Make sure branch on which leaf is to be hung exists */
  stree = traverse_branch( ConfigList, &branchCS, &node );

  if( stree == 0 ) return (-1); /* branch crosses a leaf node! */

  if( node == 0 ) { 
    ctree = create_tree( stree, &branchCS );
  } else ctree = stree;

  /*
   * Make sure LEAF does not already exist
   */
  if( find_node( ctree, &leafCS ) != 0 ) return -2;

  /* Now create leaf */
  basesize = sizeof( TreeNode ) + CSLen( leafCS ) + 1; 
  nsize    = basesize + len; /* Work out data size */ 

  if( pos == (-1) || pos >= xl_List_get_size( ctree ) ) np = ListEnd;

  xl_List_set_datasize( ctree, nsize );   /* Specify space required */
  xl_List_insert_space( ctree, np    );   /* Create node with space */
  node = (TreeNode*)xl_List_get_current_data( ctree ); /* get ptr to space*/
  
  /* Start filling in details */
  node->ttype    = T_LEAF;
  node->name     = (char*)node + sizeof( TreeNode );
  node->data     = (char*)node + basesize;
  node->dlen     = len;

  /* Add leaf name */
  memcpy( node->name, CSStart(leafCS), CSLen(leafCS) );
  node->name[ CSLen(leafCS) ] = '\0';
  node->namelen = CSLen(leafCS);

  /* Add data */    
  memcpy( node->data, data, len );
  
  node->stree = 0; /* No further nodes off this one */
  
  return 0;
}

/* ------------------------------------------------------------------------ */

int
count_children( CString *branchCS )
{
  TreeNode *node;
  CString   keepCS = CSInit;

  MirrorCString( keepCS, *branchCS );

  if( ConfigList == 0 ) return 0;

  if( traverse_branch( ConfigList, &keepCS, &node ) != 0 ) {
    
    return (node == 0 || node->stree == 0) ? 0 : xl_List_get_size(node->stree);
  }
  return 0;
}


/* ------------------------------------------------------------------------ */
/*
 * Return the list structure that the last element in a branch chain is
 * found, and also a pointer to the TreeNode of the last element if found
 * in the tree.
 *
 * Returns a pointer to the datafile recorded against the last branch node
 * it found.  This is used when writing registry to disk so that nodes off
 * branches know which data file to persist to.
 *
 * If the search fails, then the sub List in which the failure occured
 * is returned so that nodes may be added to create the required tree..
 * branchCS is returned containing the remainder of the tree that was not
 * found.
 */
LIST *
traverse_branch( LIST *tree, CString *branchCS, TreeNode** data )
{
  CString  tmpCS  = CSInit;
  CString  keepCS = CSInit;
  LIST    *stree  = tree;
  void    *node;

  MirrorCString( keepCS, (*branchCS) );

  /* 
   * Special case - when the requested branch is '/ '
   */
  if( CSLen(keepCS) == 0 || 
      ( CSLen(keepCS) == 1 && CSChar(keepCS, 0) == '/' ) ) {
    if( xl_List_get_first( stree, (uchar**)data ) == Failure ) {
      *data = 0; /* Last node not found */
    } else {
      if( (*data)->ttype != T_BRANCH ) stree = 0;
    }
    return stree;
  }
  
  while( chop_line( &tmpCS, branchCS, SEP ) == Success ) {

    if( CSLen(tmpCS) == 0 ) continue;

    /* Move down tree */
    if( (node = find_node( stree, &tmpCS )) != 0 ) { /* Found node */

      *data  = (TreeNode*)node;

      /* RETURN with NULL and node ptr if any 
       * node is NOT a branch node!! 
       */
      if( ((TreeNode*)node)->ttype != T_BRANCH ) return 0;

      if( (*data)->stree == NULL ) return 0;/*No Children!*/

      /* Move _into_ node */
      stree = (*data)->stree;

    } else {
      /* node not found */
      *data = 0; /* Last node not found */
      MirrorCString( (*branchCS), keepCS );
      break;
    }
    MirrorCString( keepCS, (*branchCS) );
  }
  return stree;
}

/* ------------------------------------------------------------------------ */
/*
 * Return pointer to requested node in branch (list)
 */
static void *
find_node( LIST *twigs, CString *nameCS )
{
  TreeNode *data    = NULL;
  TreeNode *ret     = 0;
  LIST      store   = {0};
#ifdef WILD_MATCH
  CString   nnameCS = CSInit;
#endif

  /* Keep a copy of list descriptor so we can restore it's state */
  memcpy( &store, twigs, sizeof( LIST ) );

  if( xl_List_get_first( twigs, (uchar**)&data ) == Success ) {
    do {
#ifndef WILD_MATCH
 /* #warning -= Wildcard Node Matching Disabled - Still in Development =- */
      if( data->namelen == CSLenPtr( nameCS ) &&
	  strncmp( data->name, CSStartPtr(nameCS), CSLenPtr(nameCS) ) == 0 ) {
        ret = data;
        break; /* Found */
      }
#else
#warning Wildcard Node Matching Enabled
      WrapCStringLen( nnameCS, data->name, data->namelen );
      if( wild_compare( &nnameCS, nameCS ) == 1 ) {
        ret = data;
        break; /* Found */
      }
#endif
      
    } while ( xl_List_get_next( twigs, (uchar**)&data ) == Success );
  }
  
  /* Restore list state */
  memcpy( twigs, &store, sizeof( LIST ) );

  return (void*)ret;
}

#ifdef WILD_MATCH
/* ------------------------------------------------------------------------ */
/*
 * Compares two strings.  Allows '?' matching of any character, and '*'
 * matching of everything else.  ANYTHING FOLLOWING A '*' WILL BE IGNORED!!!!
 *
 * Returns: 1 = Match
 *          0 = no match
 */
static int
wild_compare( const CString* s1, const CString* s2 )
{
  int l1 = CSLenPtr( s1 );
  int l2 = CSLenPtr( s2 );
  char * p1 = CSStartPtr( s1 );
  char * p2 = CSStartPtr( s2 );

  int max, i;
  max = ( l1 < l2 ) ? l1 : l2;

  for( i = 0; i < max; i++ ) {
    if( p1[i] != p2[i] && 
	p1[i] != '?'   && p2[i] != '?' &&
	p1[i] != '*'   && p2[i] != '*' ) {
      return 0;
    }
    
    if( p1[i] == '*' || p2[i] == '*' ) return 1; /* COMPLETE WILD MATCH! */
  }
  
  return ( l1 != l2 ) ? 0 : 1 ; /* Only success if strings same length */
}
#endif

/* ------------------------------------------------------------------------ */

void
dump_tree()
{
  traverse_tree( ConfigList, 0 );
}

/* ------------------------------------------------------------------------ */
/*
 * Used for verbose output
 */
static void
traverse_tree( LIST *tree, int tab )
{
  TreeNode *node;
  LIST      keep;

  memcpy( &keep, tree, sizeof( LIST ) );
  
  if( xl_List_get_first( tree, (uchar**)&node ) == Success ) {
    do { 
      printf( "%*.*s%c%-15.15s", tab, tab, "", (node->ttype == T_LEAF) ? '-' : '+', 
	      node->name );
      if( node->ttype == T_LEAF ) {
	printf( " %*.*s\n", (int)(node->dlen), (int)(node->dlen), node->data);
      } else printf( "\n" );
      
      if( node->ttype == T_BRANCH &&  
	  node->stree != NULL && xl_List_get_size( node->stree ) > 0 ) {
	traverse_tree( node->stree, tab + 2 );
      }
      
    } while ( xl_List_get_next( tree, (uchar**)&node ) == Success );
  }
  
  memcpy( tree, &keep, sizeof( LIST ) );
}

/* ------------------------------------------------------------------------ */

static int ind = 0;

RetValue
flatten_tree( CString *memCS )
{
  free_cstring( memCS );

  ind          = 0;
  flat_r_count = 0;

  if( flatten_tree_r( ConfigList, memCS ) < 0 ) {
    free_cstring( memCS );
    return Failure;
  }
  return Success;
}

/* ------------------------------------------------------------------------ */

static int
flatten_tree_r( LIST *tree, CString *memoryCS )
{
  TreeNode     *node;
  SHMLIST      *memlist;
  FlatTreeNode  fnode;
  CString       tmpCS    = CSInit;
  CString       ascii0CS = CSInit;
  int           os   = 0;
  int           size = 0;
  int           num_nodes = 0;
  int           tree_start_offset = -1;

  flat_r_count++; /* Recursion count */
  
  WrapCStringLen( ascii0CS, "\0", 1 );

  /* Count how many nodes we have in the tree and create space for them in
   * the flat-tree
   */
  num_nodes = xl_List_get_size(tree);
  size = shm_List_determine_memory_req( sizeof(FlatTreeNode), num_nodes );

  /* Create space for the flat list in our memory space */
  extend_cstring( memoryCS, size + BOUNDARY );
  tree_start_offset = CSLenPtr( memoryCS );

  memlist = (SHMLIST*)(CSStartPtr( memoryCS ) + tree_start_offset);

  os = 0;

  /* Align to BOUNDARY bytes word boundary */
  if( (((long)memlist) % BOUNDARY) != 0 ) {
    os = BOUNDARY - (((long)memlist) % BOUNDARY);

    tree_start_offset += os;
    memlist     += os;
  }

  /* The first time through we MUST be aligned properly, if we're not then
   * our Boundary value must be incorrect for this architecture.
   */
  if( flat_r_count == 1 && os > 0 ) {
    printf( "** SERIOUS ERROR - Memory is incorrectly aligned:\n** BOUNDARY Setting of %d is not valid for this architecture\n", BOUNDARY );
	exit(-1);
  }

  IF_DEBUG printf( "%*sCreate subtree for %d nodes\n", ind, "", num_nodes );
  IF_DEBUG printf( "%*s  Tree requires %d bytes (%d + %d)\n", ind, "", 
	  size + os, size, os );
  
  IF_DEBUG printf( "%*s  Tree start %p (boundary %d)\n", ind, "", memlist, ((int)memlist)%BOUNDARY );
  
  IF_DEBUG printf( "%*sCString before creating list-> Used: %d  Free: %d\n", ind, "", 
	  CSLenPtr(memoryCS), CSMaxPtr(memoryCS) - CSLenPtr(memoryCS) );
  
  if( shm_List_init( memlist,
		     sizeof(FlatTreeNode), num_nodes ) == Failure ) {
    return -1;
  }

  /* Step over list */
  CSLenPtr( memoryCS ) = tree_start_offset + size;

  IF_DEBUG printf( "%*sCString after creating list-> Used: %d  Free: %d\n", ind, "", 
	  CSLenPtr(memoryCS), CSMaxPtr(memoryCS) - CSLenPtr(memoryCS) );


  if( xl_List_get_first( tree, (uchar**)&node ) == Success ) {
    do { 
      
      fnode.ttype   = node->ttype;
      fnode.namelen = node->namelen;
      fnode.dlen    = node->dlen;
      
      fnode.data_o  = CSLenPtr( memoryCS );
      WrapCStringLen( tmpCS, node->data, node->dlen );
      append_cstring( memoryCS, &tmpCS    );
      append_cstring( memoryCS, &ascii0CS );  /* Null terminate data string */

      fnode.name_o  = CSLenPtr( memoryCS );
      WrapCStringLen( tmpCS, node->name, node->namelen );
      append_cstring( memoryCS, &tmpCS );

      IF_DEBUG printf( "%*sAdding node '%*.*s'\n", ind, "", node->namelen, node->namelen, node->name );
      
      if( node->ttype == T_LEAF ) fnode.stree_o = -1;
      
      if( node->ttype == T_BRANCH &&  
	  node->stree != NULL && xl_List_get_size( node->stree ) > 0 ) {

	IF_DEBUG printf( "%*sBranch node...\n", ind, "" );

	ind += 2;
	if( (os = flatten_tree_r( node->stree, memoryCS )) < 0 ) return -1;
	ind -= 2;

	fnode.stree_o = os;
	IF_DEBUG printf( "%*s...Child list had offset %d\n", ind, "", os );
      }  

      IF_DEBUG printf( "%*sInsert node '%*.*s'\n", ind, "", node->namelen, node->namelen, node->name );

      shm_List_insert_node( ((SHMLIST*)(CSStartPtr(memoryCS) + 
				       tree_start_offset)), 
			    ListEnd, (uchar*)&fnode );

      IF_DEBUG printf( "%*sInsert Done\n", ind, "" );
  
    } while ( xl_List_get_next( tree, (uchar**)&node ) == Success );
  }

  return tree_start_offset;
}


#if 0
/* ------------------------------------------------------------------------ */

static void *
find_flat_node( SHMLIST *stree, const CString *nodeCS, char *base )
{
  FlatTreeNode *fnode  = NULL;
  FlatTreeNode *ret    = NULL;

  if( shm_List_get_first( stree, (uchar**)&fnode ) == Success ) {
    do {
      if( fnode->namelen == CSLenPtr( nodeCS ) &&
	      strncmp( base + fnode->name_o, 
		           CSStartPtr(nodeCS), CSLenPtr(nodeCS) ) == 0 ) {
        ret = fnode;
		break;
      }
       
    } while( shm_List_get_next( stree, (uchar**)&fnode ) == Success );
  }

  return (void*)ret;
}


/* ------------------------------------------------------------------------ */

static SHMLIST *
traverse_flat_branch( CString *treeCS, CString *branchCS, FlatTreeNode **data )
{
  CString  tmpCS  = CSInit;
  CString  keepCS = CSInit;
  char    *base   = NULL;
  SHMLIST *stree  = NULL;
  void    *node;

  base  = CSStartPtr(treeCS);
  stree = (SHMLIST*)base;  /* Initial start tree */

  MirrorCString( keepCS, (*branchCS) );

  /* 
   * Special case - when the requested branch is '/ '
   */
  if( CSLen(keepCS) == 0 || 
      ( CSLen(keepCS) == 1 && CSChar(keepCS, 0) == '/' ) ) {
    if( shm_List_get_first( stree, (uchar**)data ) == Failure ) {
      *data = 0; /* Last node not found */
    } else {
      if( (*data)->ttype != T_BRANCH ) stree = 0;
    }
    return stree;
  }
  
  while( chop_line( &tmpCS, branchCS, '/' ) == Success ) {

    if( CSLen(tmpCS) == 0 ) continue;

    /* Move down tree */

    if( (node = find_flat_node( stree, &tmpCS, base )) != 0 ) { /* Found node */

      *data  = (FlatTreeNode*)node;

      if( ((FlatTreeNode*)node)->ttype != T_BRANCH ) return 0;
      if( (*data)->stree_o < 0 ) return 0; /* No Children! */

      /* Move _into_ node */
      stree = (SHMLIST*)(base + (*data)->stree_o);

    } else {
      /* node not found */
      *data = 0; /* Last node not found */
      MirrorCString( (*branchCS), keepCS );
      break;
    }
    MirrorCString( keepCS, (*branchCS) );
  }
  return stree;
}

/* ------------------------------------------------------------------------ */

RetValue
RegDataChar( CString *treeCS, const char *path, char **data, long *len )
{
  RetValue      ret      = Failure;
  CString       branchCS = CSInit;
  CString       leafCS   = CSInit;
  FlatTreeNode *node     = 0;
  SHMLIST      *stree    = NULL;

  /* Keep a copy of the branch */
  WrapCString( branchCS, path );
  chop_line_last( &leafCS, &branchCS, '/' ); /* Chop off leaf */
  

  /* Make sure branch on which leaf is to be hung exists */
  stree = traverse_flat_branch( treeCS, &branchCS, &node );
  
  /* Return failure if branch crosses a leaf node or branch incomplete */
  if( stree == 0 || node == 0 ) return Failure; 
  
  node = (FlatTreeNode*)find_flat_node( stree, &leafCS, CSStartPtr(treeCS) );
  
  if( node != 0 && node->ttype == T_LEAF )
  {
    *data = node->data_o + CSStartPtr(treeCS);
    *len  = node->dlen;

    ret = Success;
  }
  return ret;
}
#endif

/* ------------------------------------------------------------------------ */
/* end */
